from unittest.mock import MagicMock, mock_open, patch
from urllib.parse import urljoin
import os
import unittest

from pretend import stub
from requests import HTTPError

from gd import scrape


class Test_web_scraper(unittest.TestCase):
    """Test gd.scrape.web_scraper"""

    @patch("requests.get")
    def test_get_exception(self, mock_get):
        mock_get.side_effect = Exception
        rv = scrape.web_scraper(["lol"])
        self.assertRaises(Exception, next, rv)

    @patch("requests.get")
    def test_get_raises_status(self, mock_get):
        response = MagicMock()
        response.raise_for_status.side_effect = HTTPError
        mock_get.return_value = response
        rv = scrape.web_scraper(["lol"])
        self.assertEqual(list(rv), [])

    @patch("requests.get")
    @patch("gd.scrape.ElementTree.fromstring")
    def test_bs_raises(self, mock_fromstring, mock_get):
        mock_fromstring.side_effect = Exception
        rv = scrape.web_scraper(["lol"])
        self.assertRaises(Exception, next, rv)

    @patch("requests.get")
    @patch("gd.scrape.ElementTree.fromstring")
    def test_no_links(self, mock_fromstring, mock_get):
        source = MagicMock()
        source.findall.return_value = []
        mock_fromstring.return_value = source
        rv = scrape.web_scraper(["lol"])
        self.assertEqual(list(rv), [])

    @patch("requests.get")
    @patch("gd.scrape.ElementTree.fromstring")
    def test_matches(self, mock_fromstring, mock_get):
        root = "http://www.example.com"
        link1, link2 = MagicMock(), MagicMock()
        link1.attrib = {"href": "foo123"}
        link2.attrib = {"href": "bar456"}

        source = MagicMock()
        source.findall.return_value = [link1, link2]
        mock_fromstring.return_value = source

        for match, expected in ((None, [urljoin(root, "foo123"),
                                        urljoin(root, "bar456")]),
                                ("foo", [urljoin(root, "foo123")]),
                                ("lol", [])):
            with self.subTest(match=match):
                rv = scrape.web_scraper([root], match)
                self.assertEqual(list(rv), expected)

    def test_no_roots(self):
        expected = []
        actual = list(scrape.web_scraper([]))
        self.assertEqual(actual, expected)

    @patch("gd.scrape.ElementTree.fromstring")
    def test_requests_session(self, mock_fromstring):
        response = stub(raise_for_status=lambda: None, content="content")
        session = stub(get=lambda arg: response)

        root = "http://www.example.com"
        source = MagicMock()
        link1 = MagicMock()
        link1.attrib = {"href": "foo123"}
        source.findall.return_value = [link1]
        mock_fromstring.return_value = source

        expected = [urljoin(root, "foo123")]
        rv = scrape.web_scraper([root], session=session)
        self.assertEqual(list(rv), expected)


class Test_filesystem_scraper(unittest.TestCase):
    """Test gd.scrape.filesystem_scraper"""

    @patch("os.listdir")
    def test_scraper(self, mock_listdir):
        root = "rootdir"
        file1, file2 = "foo123", "bar456"
        mock_listdir.return_value = [file1, file2]

        for match, expected in ((None, [os.path.join(root, file1),
                                        os.path.join(root, file2)]),
                                ("foo", [os.path.join(root, file1)]),
                                ("lol", [])):
            with self.subTest(match=match):
                rv = scrape.filesystem_scraper([root], match)
                self.assertEqual(list(rv), expected)

    def test_no_roots(self):
        expected = []
        actual = list(scrape.filesystem_scraper([]))
        self.assertEqual(actual, expected)


class Test_get_urls(unittest.TestCase):
    """Test all gd.scrape.get_* functions"""

    def test_get(self):
        expected = "abcdefg"

        def fake_scraper(*args):
            yield from expected

        for fn in (scrape.get_years, scrape.get_months, scrape.get_days,
                   scrape.get_games):
            with self.subTest(function=fn):
                actual = fn(["root"], source=fake_scraper)
                self.assertEqual(list(actual), list(expected))

    def test_get_files(self):
        expected = ["players.xml",
                    "game.xml",
                    "inning_all.xml"]

        def fake_scraper(root, match, session):
            """Just yield back the match, make sure we end up with
            the same amount of items."""
            yield from [match]

        actual = scrape.get_files(["root"], source=fake_scraper)
        self.assertEqual(list(actual), list(expected))


class Test_download(unittest.TestCase):
    """Test gd.scrape.download"""

    @patch("requests.Session.get")
    @patch("os.makedirs")
    def test_file(self, mock_makedirs, mock_get):
        urls = ["http://gd.mlb.com/test1.xml", "http://gd.mlb.com/test2.xml"]
        targets = ["gd.mlb.com/test1.xml", "gd.mlb.com/test1.xml"]
        content = "content"

        response = MagicMock()
        response.raise_for_status = MagicMock()
        response.content.side_effect = content
        mock_get.return_value = response

        mo = mock_open()
        # Need to patch `open` within the gd.scrape namespace.
        with patch("%s.open" % scrape.__name__, mo, create=True):
            scrape.download(urls)

        mo.assert_any_call(targets[0], "w")
        mo.assert_any_call(targets[1], "w")

    @patch("requests.Session.get")
    @patch("os.makedirs")
    def test_directory(self, mock_makedirs, mock_get):
        url = "http://gd.mlb.com/inning/"

        response = MagicMock()
        response.raise_for_status = MagicMock()
        mock_get.return_value = response

        mo = mock_open()
        # Need to patch `open` within the gd.scrape namespace.
        with patch("%s.open" % scrape.__name__, mo, create=True):
            scrape.download([url])

    @patch("requests.Session.get")
    def test_HTTPError(self, mock_get):
        url = "http://gd.mlb.com/inning/inning_all.xml"

        response = MagicMock()
        response.raise_for_status = MagicMock(side_effect=HTTPError)
        mock_get.return_value = response

        scrape.download([url])



from datetime import datetime, date, time
import unittest

from pretend import stub

from gd import utils


class Test_get_boundary(unittest.TestCase):

    def test_with_namedtuple(self):
        actual = utils.get_boundary("")
        self.assertIsNone(actual.date)
        self.assertEqual(actual.num_parts, 0)

    def test_None(self):
        actual_dt, actual_parts = utils.get_boundary(None)
        self.assertIsNone(actual_dt)
        self.assertEqual(actual_parts, 0)

    def test_empty(self):
        actual_dt, actual_parts = utils.get_boundary("")
        self.assertIsNone(actual_dt)
        self.assertEqual(actual_parts, 0)

    def test_bad_data(self):
        actual_dt, actual_parts = utils.get_boundary("lololololol")
        self.assertIsNone(actual_dt)
        self.assertEqual(actual_parts, 0)

    def test_year(self):
        year = "2014"
        expected_parts = 1
        expected_dt = datetime.strptime(year, "%Y")
        actual_dt, actual_parts = utils.get_boundary(year)
        self.assertEqual(actual_dt, expected_dt)
        self.assertEqual(actual_parts, expected_parts)

    def test_year_month(self):
        year = "2014-07"
        expected_parts = 2
        expected_dt = datetime.strptime(year, "%Y-%m")
        actual_dt, actual_parts = utils.get_boundary(year)
        self.assertEqual(actual_dt, expected_dt)
        self.assertEqual(actual_parts, expected_parts)

    def test_year_month_day(self):
        year = "2014-07-02"
        expected_parts = 3
        expected_dt = datetime.strptime(year, "%Y-%m-%d")
        actual_dt, actual_parts = utils.get_boundary(year)
        self.assertEqual(actual_dt, expected_dt)
        self.assertEqual(actual_parts, expected_parts)


class Test_get_inclusive_urls(unittest.TestCase):

    def _test_range(self, start, stop, expected):
        urls = ["a/a/a", "a/b/a", "a/c/a", "a/d/a", "a/e/a"]
        actual = utils.get_inclusive_urls(urls, start, stop)
        self.assertEqual(list(actual), expected)

    def test_early_range(self):
        start = "a/a"
        stop = "a/c"
        expected = ["a/a/a", "a/b/a", "a/c/a"]
        self._test_range(start, stop, expected)

    def test_middle_range(self):
        start = "a/b"
        stop = "a/d"
        expected = ["a/b/a", "a/c/a", "a/d/a"]
        self._test_range(start, stop, expected)

    def test_late_range(self):
        start = "a/c"
        stop = "a/e"
        expected = ["a/c/a", "a/d/a", "a/e/a"]
        self._test_range(start, stop, expected)

    def test_no_matcing_range(self):
        start = "a/f"
        stop = "a/g"
        expected = []
        self._test_range(start, stop, expected)


class Test_datetime_to_url(unittest.TestCase):
    """Test gd.utils.datetime_to_url"""

    def test_dt(self):
        expected = "year_1984/month_07/day_02/"
        dt = stub(year=1984, month=7, day=2)
        actual = utils.datetime_to_url(dt)
        self.assertEqual(actual, expected)


class Test_create_date(unittest.TestCase):
    """Test gd.utils.create_date"""

    def test_create(self):
        actual = utils.create_date("July 2, 1984")
        self.assertEqual(actual, date(1984, 7, 2))

    def test_create_empty(self):
        actual = utils.create_date("")
        self.assertEqual(actual, date.min)


class Test_create_datetime(unittest.TestCase):
    """Test gd.utils.create_datetime"""

    def test_create_datetime(self):
        actual = utils.create_datetime("1984-07-02T12:34:56Z")
        self.assertEqual(actual, datetime(1984, 7, 2, 12, 34, 56))

    def test_create_datetime_empty(self):
        actual = utils.create_datetime("")
        self.assertEqual(actual, datetime.min)


class Test_create_time(unittest.TestCase):
    """Test gd.utils.create_time"""

    def test_create_time1(self):
        actual = utils.create_time("123456")
        self.assertEqual(actual, time(12, 34))

    def test_create_time2(self):
        actual = utils.create_time("12:34")
        self.assertEqual(actual, time(12, 34))

    def test_create_time3(self):
        actual = utils.create_time("1234")
        self.assertEqual(actual, time(12, 34))

    def test_create_time_empty(self):
        actual = utils.create_time("")
        self.assertEqual(actual, time.min)



"""
Parser for MLB's Gameday XML data
"""
from gd.utils import create_date, create_datetime, create_time


def _strip_keys(obj, valid):
    """Given a dictionary and a set of valid keys, return a dictionary which
    only contains keys found in `valid`."""
    return {key: obj[key] for key in obj if key in valid}


def _get_all_attributes(tree, element, valid):
    """Given a tree, yield the attribute dictionary of each element."""
    yield from [_strip_keys(x.attrib, valid) for x in tree.findall(element)]


def get_date(tree):
    """Parse players.xml file and return the date these players are playing."""
    return create_date(tree.attrib["date"])


def get_players(tree):
    """Parse players.xml file and return players and their attributes."""
    valid_keys = ("id", "first", "last", "num", "boxname", "rl", "bats",
                  "position", "status", "team_id")
    yield from _get_all_attributes(tree, ".//player", valid_keys)


def get_plate_umpire(tree):
    """Parse players.xml data to find the plate umpire."""
    umpires = tree.findall(".//umpire")
    for ump in umpires:
        if ump.attrib["position"] == "home":
            return {k: v for k, v in ump.attrib.items() if k in ("id", "name")}
    else:
        raise Exception("No plate umpire found.")


def get_game(tree):
    """Parse game.xml data to find the game information."""
    valid_keys = ("game_pk", "type", "local_game_time", "game_time_et",
                  "gameday_sw", "home_team", "away_team", "stadium",
                  "plate_umpire")
    tree.attrib["local_game_time"] = create_time(
        tree.attrib["local_game_time"])
    return {k: v for k, v in tree.attrib.items() if k in valid_keys}


def get_actions(tree):
    valid_keys = ("id", "game_pk", "b", "s", "o", "pitch", "player", "event",
                  "event2", "des_es", "des", "tfs", "tfs_zulu")

    for action in tree.findall(".//action"):
        action.attrib["tfs"] = create_time(action.attrib["tfs"])
        action.attrib["tfs_zulu"] = create_datetime(action.attrib["tfs_zulu"])
        yield {k: v for k, v in action.attrib.items() if k in valid_keys}


def get_teams(tree):
    """Parse game.xml data to find the teams involved."""
    valid_keys = ("code", "id", "name", "name_full", "name_brief",
                  "division_id", "league_id", "league")
    yield from _get_all_attributes(tree, ".//team", valid_keys)


def get_stadium(tree):
    """Parse game.xml data to find the stadium."""
    stadium = tree.find(".//stadium")
    if stadium is None:
        raise Exception("Did not find a stadium.")

    valid_keys = ("id", "name", "location")

    return {k: v for k, v in stadium.attrib.items() if k in valid_keys}


def get_atbats(tree):
    """Parse inning_all.xml data to find the atbats."""
    valid_keys = ("num", "b", "s", "o", "start_tfs", "start_tfs_zulu",
                  "batter", "stand", "b_height", "pitcher", "p_throws",
                  "des", "des_es", "event", "score", "home_team_runs",
                  "away_team_runs")
    for atbat in _get_all_attributes(tree, ".//atbat", valid_keys):
        atbat["start_tfs"] = create_time(atbat["start_tfs"])
        atbat["start_tfs_zulu"] = create_datetime(
            atbat["start_tfs_zulu"])
        if "score" in atbat:
            atbat["score"] = True if atbat["score"] == "T" else False
        else:
            atbat["score"] = False

        yield atbat


def get_pitches(tree):
    """Parse inning_all.xml data to find the pitches."""
    for atbat in tree.findall(".//atbat"):
        for pitch in atbat.findall(".//pitch"):
            pitch.attrib["tfs"] = create_time(pitch.attrib["tfs"])
            pitch.attrib["tfs_zulu"] = create_datetime(
                pitch.attrib["tfs_zulu"])
            # Tie pitches back to the atbat they came from.
            pitch.attrib["start_tfs_zulu"] = create_datetime(
                atbat.attrib["start_tfs_zulu"])
            yield pitch.attrib



from datetime import datetime, date, time
import unittest

from pretend import stub

from gd import parser


class Test_get_game(unittest.TestCase):
    """Test the gd.parser.get_game function."""

    def test_get_game(self):
        expected = {"local_game_time": "12:00"}
        tree = stub(attrib=expected)
        actual = parser.get_game(tree)
        self.assertEqual(actual, expected)


class Test_get_date(unittest.TestCase):
    """Test the gd.parser.get_date function."""

    def test_get_date(self):
        game = {"date": "July 2, 1984"}
        tree = stub(attrib=game)
        actual = parser.get_date(tree)
        self.assertEqual(actual, date(1984, 7, 2))


class Test_get_players(unittest.TestCase):
    """Test the gd.parser.get_players function."""

    def test_get_players(self):
        expected = {"first": "Ryne", "last": "Sandberg"}
        s1 = stub(attrib=expected)
        s2 = stub(attrib=expected)
        s3 = stub(attrib=expected)
        tree = stub(findall=lambda arg: [s1, s2, s3])
        actual = parser.get_players(tree)
        self.assertEqual(list(actual), [expected]*3)


class Test_get_plate_umpire(unittest.TestCase):
    """Test the gd.parser.get_plate_umpire function."""

    def test_get_plate_umpire(self):
        attrib = {"position": "home", "name": "Gerry Davis"}
        expected = {"name": "Gerry Davis"}
        element = stub(attrib=attrib)
        tree = stub(findall=lambda arg: [element])
        actual = parser.get_plate_umpire(tree)
        self.assertEqual(actual, expected)

    def test_get_plate_umpire_no_pu(self):
        attrib = {"position": "first", "name": "Ted Barrett"}
        element = stub(attrib=attrib)
        tree = stub(findall=lambda arg: [element])
        self.assertRaisesRegex(Exception, "No plate umpire found.",
                               parser.get_plate_umpire, tree)

    def test_get_plate_umpire_no_umpires(self):
        tree = stub(findall=lambda arg: [])
        self.assertRaisesRegex(Exception, "No plate umpire found.",
                               parser.get_plate_umpire, tree)


class Test_get_teams(unittest.TestCase):
    """Test the gd.parser.get_teams function."""

    def test_get_teams(self):
        value = {"name": "Chicago Cubs"}
        team = stub(attrib=value)
        tree = stub(findall=lambda arg: [team, team])
        actual = parser.get_teams(tree)
        self.assertEqual(list(actual), [value, value])


class Test_get_stadium(unittest.TestCase):
    """Test the gd.parser.get_stadium function."""

    def test_get_stadium(self):
        expected = {"name": "Wrigley Field"}
        stadium = stub(attrib=expected)
        tree = stub(find=lambda arg: stadium)
        actual = parser.get_stadium(tree)
        self.assertEqual(actual, expected)

    def test_get_stadium_missing(self):
        tree = stub(find=lambda arg: None)
        self.assertRaisesRegex(Exception, "Did not find a stadium.",
                               parser.get_stadium, tree)


class Test_get_actions(unittest.TestCase):
    """Test the gd.parser.get_actions function."""

    def test_get_actions(self):
        input_value = {"tfs": "123456",
                       "tfs_zulu": "2014-07-19T23:12:35Z",
                       "id": "987654"}
        value = stub(attrib=input_value)
        tree = stub(findall=lambda arg: [value])
        actual = parser.get_actions(tree)

        expected = {"tfs": time(12, 34),
                    "tfs_zulu": datetime(2014, 7, 19, 23, 12, 35),
                    "id": "987654"}
        self.assertEqual(list(actual), [expected])


class Test_get_atbats(unittest.TestCase):
    """Test the gd.parser.get_atbats function."""

    def test_get_atbats_no_atbats(self):
        tree = stub(findall=lambda arg: [])
        actual = parser.get_atbats(tree)
        self.assertEqual(list(actual), [])

    def test_get_atbats(self):
        input_value = {"start_tfs": "123456",
                       "start_tfs_zulu": "2014-07-19T23:12:35Z",
                       "score": "T"}
        value = stub(attrib=input_value)
        atbat = stub(findall=lambda arg: [value],
                     attrib=input_value)
        tree = stub(findall=lambda arg: [atbat])

        actual = parser.get_atbats(tree)
        expected = {"start_tfs": time(12, 34),
                    "start_tfs_zulu": datetime(2014, 7, 19, 23, 12, 35),
                    "score": True}
        self.assertEqual(list(actual), [expected])


class Test_get_pitches(unittest.TestCase):
    """Test the gd.parser.get_pitches function."""

    def test_get_pitches(self):
        atbat_attribs = {"start_tfs_zulu": "2014-07-19T23:12:35Z"}
        pitch_attribs = {"tfs": "123456",
                         "tfs_zulu": "2014-07-19T23:15:35Z"}
        pitch_value = stub(attrib=pitch_attribs)

        atbat = stub(findall=lambda arg: [pitch_value],
                     attrib=atbat_attribs)
        tree = stub(findall=lambda arg: [atbat])

        actual = parser.get_pitches(tree)
        expected = {"tfs": time(12, 34),
                    "tfs_zulu": datetime(2014, 7, 19, 23, 15, 35),
                    "start_tfs_zulu": datetime(2014, 7, 19, 23, 12, 35)}
        self.assertEqual(list(actual), [expected])



from sqlalchemy.orm import relationship, backref
from sqlalchemy import (Column, Date, DateTime, Time, ForeignKey, Integer,
                        String, Numeric, Boolean)

from gd.database import Base


class Team(Base):
    __tablename__ = "team"

    code = Column(String, primary_key=True)
    id = Column(Integer, primary_key=True)
    name = Column(String, nullable=False)
    name_full = Column(String, nullable=False)
    name_brief = Column(String, nullable=False)
    division_id = Column(Integer, nullable=False)
    league_id = Column(Integer, nullable=False)
    league = Column(String, nullable=False)

    players = relationship("Player")
    replays = relationship("Replay", backref="team")

    def __repr__(self):
        return "<%s id=%s>" % (self.name_full, self.id)


class Player(Base):
    __tablename__ = "player"

    id = Column(Integer, nullable=False, primary_key=True)
    first = Column(String, nullable=False)
    last = Column(String, nullable=False)
    num = Column(Integer, nullable=False)
    boxname = Column(String, nullable=False)
    rl = Column(String, nullable=False)
    bats = Column(String, nullable=False)
    position = Column(String, nullable=False)
    status = Column(String, nullable=False)
    team_id = Column(Integer, ForeignKey("team.id"), nullable=False)

    # The following is available in the /batters/ dir. Could be merged here.
    # height  = Column(String, nullable=False)
    # weight = Column(Integer, nullable=False)
    # throws = Column(String, nullable=False)
    # dob = Column(String, nullable=False)

    def __repr__(self):
        return "<Player id=%s, name=%s %s>" % (self.id, self.first, self.last)


class Umpire(Base):
    __tablename__ = "umpire"

    id = Column(Integer, nullable=False, primary_key=True)
    name = Column(String, nullable=False)

    games = relationship("Game")

    def __repr__(self):
        return "Umpire(id=%d, name='%s')" % (self.id, self.name)


class Stadium(Base):
    __tablename__ = "stadium"

    id = Column(Integer, nullable=False, primary_key=True)
    name = Column(String, nullable=False)
    location = Column(String, nullable=False)

    def __repr__(self):
        return "Stadium(id=%d, name='%s', location='%s')" % (self.id,
                                                             self.name,
                                                             self.location)


class Game(Base):
    __tablename__ = "game"

    game_pk = Column(Integer, nullable=False, primary_key=True)
    type = Column(String, nullable=False)
    date = Column(Date, nullable=False)
    local_game_time = Column(Time, nullable=False)
    game_time_et = Column(String, nullable=False)
    gameday_sw = Column(String, nullable=False)
    home_team = Column(Integer, ForeignKey("team.code"), nullable=False)
    away_team = Column(Integer, ForeignKey("team.code"), nullable=False)
    stadium = Column(Integer, ForeignKey("stadium.id"), nullable=False)
    umpire_id = Column(Integer, ForeignKey("umpire.id"), nullable=False)

    atbats = relationship("AtBat", backref="game")
    actions = relationship("Action", backref="game")
    pitches = relationship("Pitch", backref="game")
    replays = relationship("Replay", backref="game")
    umpire = relationship("Umpire", backref=backref("game", uselist=False))

    def __repr__(self):
        return "Game #%d %s@%s" % (self.game_pk, self.away_team,
                                   self.home_team)


class AtBat(Base):
    __tablename__ = "atbat"

    atbat_id = Column(Integer, autoincrement=True, primary_key=True)
    num = Column(Integer, nullable=False)
    b = Column(Integer, nullable=False)
    s = Column(Integer, nullable=False)
    o = Column(Integer, nullable=False)
    start_tfs = Column(Time, nullable=False)
    start_tfs_zulu = Column(DateTime, nullable=False)
    batter = Column(Integer, ForeignKey("player.id"), nullable=False)
    stand = Column(String, nullable=False)
    b_height = Column(String, nullable=False)
    pitcher = Column(Integer, ForeignKey("player.id"), nullable=False)
    p_throws = Column(String, nullable=False)
    des = Column(String, nullable=False)
    des_es = Column(String, nullable=False)
    event = Column(String, nullable=False)
    score = Column(Boolean)
    home_team_runs = Column(Integer)
    away_team_runs = Column(Integer)

    # Added
    game_pk = Column(Integer, ForeignKey("game.game_pk"), nullable=False)

    replay = relationship("Replay", backref="atbat")
    pitches = relationship("Pitch", backref="atbat")

    def __repr__(self):
        return "AtBat #%d, batter %d facing pitcher %d" % (self.num,
                                                           self.batter,
                                                           self.pitcher)


class Pitch(Base):
    __tablename__ = "pitch"

    pitch_id = Column(Integer, autoincrement=True, primary_key=True)
    des = Column(String, nullable=False)
    des_es = Column(String, nullable=False)
    # This is an in-game pitch ID, not universal.
    id = Column(Integer, nullable=False)
    type = Column(String, nullable=False)
    tfs = Column(Time)
    tfs_zulu = Column(DateTime)
    x = Column(Numeric)
    y = Column(Numeric)
    on_1b = Column(Integer)
    on_2b = Column(Integer)
    on_3b = Column(Integer)
    sv_id = Column(String)
    start_speed = Column(Numeric)
    end_speed = Column(Numeric)
    sz_top = Column(Numeric)
    sz_bot = Column(Numeric)
    pfx_x = Column(Numeric)
    pfx_z = Column(Numeric)
    px = Column(Numeric)
    pz = Column(Numeric)
    x0 = Column(Numeric)
    y0 = Column(Numeric)
    z0 = Column(Numeric)
    vx0 = Column(Numeric)
    vy0 = Column(Numeric)
    vz0 = Column(Numeric)
    ax = Column(Numeric)
    ay = Column(Numeric)
    az = Column(Numeric)
    break_y = Column(Numeric)
    break_angle = Column(Numeric)
    break_length = Column(Numeric)
    pitch_type = Column(String)
    type_confidence = Column(Numeric)
    zone = Column(Integer)
    nasty = Column(Integer)
    spin_dir = Column(Numeric)
    spin_rate = Column(Numeric)
    cc = Column(String)
    mt = Column(String)

    # Added
    start_tfs_zulu = Column(DateTime, ForeignKey("atbat.start_tfs_zulu"),
                            nullable=False)
    game_pk = Column(Integer, ForeignKey("game.game_pk"), nullable=False)

    def __repr__(self):
        return "<Pitch id=%s at %s>" % (self.id, self.tfs_zulu)


class Action(Base):
    __tablename__ = "action"

    id = Column(Integer, autoincrement=True, primary_key=True)
    game_pk = Column(Integer, ForeignKey("game.game_pk"), nullable=False)
    b = Column(Integer, nullable=False)
    s = Column(Integer, nullable=False)
    o = Column(Integer, nullable=False)
    pitch = Column(Integer, nullable=False)
    player = Column(Integer, nullable=False)
    event = Column(String, nullable=False)
    event2 = Column(String)
    des = Column(String, nullable=False)
    des_es = Column(String)
    tfs = Column(Time, nullable=False)
    tfs_zulu = Column(DateTime, nullable=False)

    replay = relationship("Replay", backref="action")

    def __repr__(self):
        return "<Action id=%s %s>" % (self.id, self.event)


class Replay(Base):
    __tablename__ = "replay"

    id = Column(Integer, autoincrement=True, primary_key=True)
    atbat_id = Column(Integer, ForeignKey("atbat.atbat_id"))
    action_id = Column(Integer, ForeignKey("action.id"))
    team_id = Column(Integer, ForeignKey("team.id"))
    game_pk = Column(Integer, ForeignKey("game.game_pk"), nullable=False)
    play = Column(String, nullable=False)
    result = Column(String, nullable=False)
    des = Column(String, nullable=False)
    des_es = Column(String)
    tfs = Column(Time, nullable=False)
    tfs_zulu = Column(DateTime, nullable=False)

    def __repr__(self):
        return "<Replay id=%s result=%s>" % (self.id, self.result)



from collections import namedtuple
from datetime import date, datetime, time
from logging.handlers import RotatingFileHandler
import logging
import os


Boundary = namedtuple("Boundary", "date num_parts")

MONTHS = {"January": 1, "February": 2, "March": 3, "April": 4,
          "May": 5, "June": 6, "July": 7, "August": 8, "September": 9,
          "October": 10, "November": 11, "December": 12}


def get_boundary(date):
    """Format a boundary date string and return the datetime and the number
    of parts the datetime object was constructed from.

    This could be a partial date consisting of a year;
    year and month; or year, month, and day."""
    parts = 0
    for fmt in ("%Y", "%Y-%m", "%Y-%m-%d"):
        try:
            parts += 1
            return Boundary(datetime.strptime(date, fmt), parts)
        except (TypeError, ValueError):
            continue
    else:
        return Boundary(None, 0)


def get_inclusive_urls(urls, start, stop):
    """Yield URLs which are of the range [start, stop]"""
    in_range = False
    out_range = False
    for url in urls:
        # Check both that a URL contains or is contained within one of
        # the boundaries. This is necessary as deeper links come.
        if url in start or start in url:
            in_range = True
        if url in stop or stop in url:
            out_range = True
        if in_range:
            yield url
        if out_range:
            break


def datetime_to_url(dt, parts=3):
    """Convert a Python datetime into the date portion of a Gameday URL"""
    fragments = ["year_{0.year:04}", "month_{0.month:02}", "day_{0.day:02}"]
    return "/".join(fragments[:parts]).format(dt) + "/"


def create_date(string):
    """Given a string like 'July 2, 1984', turn it into a datetime.date."""
    if not string:
        return date.min
    parts = string.split()
    month = MONTHS[parts[0]]
    day = int(parts[1][:-1])
    year = int(parts[2])
    return date(year, month, day)


def create_datetime(string):
    """Given a string like 2014-07-19T23:12:35Z, return a datetime.datetime."""
    if not string:
        # There are times where pitches have had an empty string for
        # a timestamp, so return the minimum value.
        return datetime.min
    return datetime.strptime(string, "%Y-%m-%dT%H:%M:%SZ")


def create_time(string):
    """Given a string like 12:00 or 1200, return a datetime.time."""
    if not string:
        return time.min
    string = string.replace(":", "")
    hour = int(string[:2])
    minute = int(string[2:4])
    return time(hour, minute)


def get_logger(name):
    logger = logging.getLogger(name)
    logger.addHandler(logging.NullHandler())
    return logger


def enable_logging():
    logger = get_logger("gd")

    level = logging.DEBUG if os.getenv("DEBUG", False) else logging.INFO
    logger.setLevel(level)
    fmt = logging.Formatter(
        "%(asctime)s | %(name)s | %(levelname)s | %(message)s")

    ch = logging.StreamHandler()
    ch.setLevel(level)
    ch.setFormatter(fmt)
    logger.addHandler(ch)

    fh = RotatingFileHandler("gd-util.log", maxBytes=100000, backupCount=5)
    fh.setLevel(level)
    fh.setFormatter(fmt)
    logger.addHandler(fh)

    return logger



from urllib.parse import urljoin, urlsplit
from xml.etree import ElementTree
import os

import requests

from gd import storage
from gd import utils

logger = utils.get_logger(__name__)

WEB_ROOT = "http://gd2.mlb.com/components/game/mlb/"

# ElementTree chokes on the doctype in these files so just skip over it.
# We're using an XML parser to parse HTML. Whatever, the rest of it works.
WITHOUT_DOCTYPE = slice(56, -1)


def upload(urls):
    session = requests.Session()

    driver = storage.get_driver()
    container = None
    uploads = 0
    for url in urls:
        parts = urlsplit(url)
        filename = os.path.split(parts.path)[1]
        # Skip directory pages.
        if not filename:
            continue

        if container is None:
            container = storage.get_container(driver, parts.netloc)

        object_name = parts.path

        response = session.get(url)
        try:
            response.raise_for_status()
        except requests.HTTPError as exc:
            logger.error("%s upload failed: %s", url, exc)
            continue

        storage.upload_object(driver, container, object_name, response.content)
        uploads += 1
    logger.info("Uploaded %d objects", uploads)


def download(urls):
    """Download `urls` into `root`. Return the count of files downloaded.
    Each URL is stored as its full URL (minus the scheme)."""
    session = requests.Session()
    downloads = 0
    for url in urls:
        parts = urlsplit(url)
        directory, filename = os.path.split(parts.path)
        # Skip directory pages.
        if not filename:
            continue

        target = parts.netloc + directory
        # Ignore if the target directory already existed.
        os.makedirs(target, exist_ok=True)

        response = session.get(url)
        try:
            response.raise_for_status()
        except requests.HTTPError as exc:
            logger.error("download error: %s raised %s", url, str(exc))
            continue

        with open(os.path.join(target, filename), "w") as fh:
            fh.write(response.content.decode("utf8"))
            logger.debug("downloaded %s", url)
            downloads += 1

    logger.info("Downloaded %d files" % downloads)


def web_scraper(roots, match=None, session=None):
    """Yield URLs in a directory which start with `match`.
    If `match` is None, all links are yielded."""
    for root in roots:
        if session is not None:
            response = session.get(root)
        else:
            response = requests.get(root)
        try:
            response.raise_for_status()
        except requests.HTTPError as exc:
            logger.error("scraping %s raised %s", root, str(exc))
            continue

        # Parse the directory listing, but ignore the DOCTYPE.
        source = ElementTree.fromstring(response.content[WITHOUT_DOCTYPE])
        a_tags = source.findall(".//a")
        for a in a_tags:
            url = a.attrib["href"]
            if match is None or url[slice(0, len(match))] == match:
                yield urljoin(root, url)


def filesystem_scraper(roots, match=None, **kwargs):
    """Yield paths in a directory which start with `match`.
    If `match` is None, all files are yielded."""
    for root in roots:
        for name in os.listdir(root):
            if match is None or name.startswith(match):
                yield os.path.join(root, name)


def get_years(root=WEB_ROOT, source=web_scraper, session=None):
    """From the root URL, yield URLs to the available years."""
    yield from source([root], "year", session)


def get_months(years, source=web_scraper, session=None):
    """Yield URLs to the available months for every year."""
    yield from source(years, "month", session)


def get_days(months, source=web_scraper, session=None):
    """Yield URLs to the available days for every month."""
    yield from source(months, "day", session)


def get_games(days, source=web_scraper, session=None):
    """Yield URLs to the available games for every day."""
    yield from source(days, "gid", session)


def get_files(games, source=web_scraper, session=None):
    """Yield URLs to the relevant files for every game."""
    for game in games:
        yield from source([game], "players.xml", session)
        yield from source([game], "game.xml", session)
        yield from source([urljoin(game, "inning/")],
                          "inning_all.xml", session)



#! /usr/bin/env python

from setuptools import setup

setup(name="gameday",
      version="0.0.0",
      description="Tools for working with MLB's Gameday data",
      author="Brian Curtin",
      author_email="brian@python.org",
      packages=["gd",
                "gd.scripts"],
      scripts=["bin/gd-util"],
      data_files=[("/usr/local/etc/gd", ["config/db.conf.example",
                                         "config/storage.conf.example"])],
      install_requires=["requests",
                        "sqlalchemy"],
      test_suite="gd.tests",
      classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3 :: Only"]
     )



import configparser
import io

from libcloud.storage.providers import get_driver as get_storage_driver
from libcloud.storage.types import ContainerDoesNotExistError


def get_driver(config="/usr/local/etc/gd/storage.conf"):
    parser = configparser.ConfigParser()
    if parser.read(config):
        provider = parser.get("storage", "provider")
        username = parser.get("storage", "username")
        api_key = parser.get("storage", "api_key")
        region = parser.get("storage", "region")
    else:
        raise Exception("Unable to read storage configuration")

    Driver = get_storage_driver(provider)
    return Driver(username, api_key, region=region)


def get_container(driver, container_name):
    """Get or create a container."""
    try:
        return driver.get_container(container_name)
    except ContainerDoesNotExistError:
        return driver.create_container(container_name)


def upload_object(driver, container, object_name, data):
    """Upload an object through libcloud.storage's upload_object_via_stream."""

    if not getattr(data, "__next__", False):
        data = io.BytesIO(data)

    return driver.upload_object_via_stream(data, container, object_name)



from datetime import datetime, timedelta
from os.path import join
from urllib.parse import urljoin
from xml.etree import ElementTree
import argparse
import os

import requests

from gd import database
from gd import parser
from gd import scrape
from gd import utils
from gd.models import (Action, AtBat, Game, Player, Pitch, Stadium,
                       Team, Umpire)

logger = utils.enable_logging()

ROOT = "gd2.mlb.com/components/game/mlb/year_2014/"


def _get_request_range(begin, end):
    if begin.date is None:
        start = scrape.WEB_ROOT
    else:
        start = urljoin(scrape.WEB_ROOT,
                        utils.datetime_to_url(begin.date, begin.num_parts))

    if end.date is None:
        stop = urljoin(scrape.WEB_ROOT, utils.datetime_to_url(
                       datetime.today() - timedelta(days=1)))
    else:
        stop = urljoin(scrape.WEB_ROOT,
                       utils.datetime_to_url(end.date, end.num_parts))

    return start, stop


def _get_file_list(session, begin, end):
    all_years = scrape.get_years(session=session)
    inc_years = utils.get_inclusive_urls(all_years, begin, end)

    all_months = scrape.get_months(inc_years, session=session)
    inc_months = utils.get_inclusive_urls(all_months, begin, end)

    all_days = scrape.get_days(inc_months, session=session)
    inc_days = utils.get_inclusive_urls(all_days, begin, end)

    games = scrape.get_games(inc_days, session=session)
    files = scrape.get_files(games, session=session)

    return files


def do_scrape(args):
    """Run the scraper over the range [begin, end]

    If no beginning is given, scraping starts from the root.
    If no ending is given, scraping ends at the yesterday's date.
    Note: end=yesterday because the schedule is pre-loaded, so scraping
    for today would mean having to account for a game existing but no files
    available."""
    if not any([args.download, args.upload]):
        print("Must choose to upload or download.")
        return -1
    action = scrape.download if args.download else scrape.upload

    start_scrape = datetime.now()
    begin = utils.get_boundary(args.begin)
    end = utils.get_boundary(args.end)

    session = requests.Session()

    start, stop = _get_request_range(begin, end)
    files = _get_file_list(session, start, stop)

    action(files)
    end_scrape = datetime.now()
    logger.debug("%s completed in %s", str(action),
                 str(end_scrape - start_scrape))


def add_teams(session, teams):
    query = session.query(Team.id)
    for team in teams:
        if query.filter(Team.id.is_(int(team["id"]))).scalar() is None:
            session.add(Team(**team))


def add_players(session, players):
    query = session.query(Player.id)
    for player in players:
        if query.filter(Player.id.is_(int(player["id"]))).scalar() is None:
            session.add(Player(**player))


def add_umpire(session, umpire):
    query = session.query(Umpire.id)
    if query.filter(Umpire.id.is_(int(umpire["id"]))).scalar() is None:
        session.add(Umpire(**umpire))


def add_stadium(session, stadium):
    query = session.query(Stadium.id)
    if query.filter(Stadium.id.is_(int(stadium["id"]))).scalar() is None:
        session.add(Stadium(**stadium))


def add_game(session, game):
    query = session.query(Game.game_pk)
    if query.filter(Game.game_pk.is_(int(game["game_pk"]))).scalar() is None:
        session.add(Game(**game))


def do_initdb(args):
    database.init()


def do_import(args):
    for root, dirs, files in os.walk(args.root):
        game_path = join(root, "game.xml")
        inning_path = join(root, "inning/inning_all.xml")
        player_path = join(root, "players.xml")

        exist = [os.path.exists(x) for x in [game_path, inning_path,
                                             player_path]]
        if not all(exist):
            continue

        with open(game_path, "r") as game_file:
            game_tree = ElementTree.fromstring(game_file.read())
        with open(inning_path, "r") as inning_file:
            inning_tree = ElementTree.fromstring(inning_file.read())
        with open(player_path, "r") as player_file:
            player_tree = ElementTree.fromstring(player_file.read())

        date = parser.get_date(player_tree)
        teams = list(parser.get_teams(game_tree))
        players = parser.get_players(player_tree)
        plate_umpire = parser.get_plate_umpire(player_tree)
        stadium = parser.get_stadium(game_tree)
        game = parser.get_game(game_tree)
        atbats = list(parser.get_atbats(inning_tree))
        pitches = list(parser.get_pitches(inning_tree))
        actions = list(parser.get_actions(inning_tree))

        # Skip spring training and exhibition games since they won't have
        # any data, and I've also seen players in these games with non-unique
        # player IDs. Just forget that...
        if game["type"] in ("S", "E"):
            continue

        add_teams(database.session, teams)
        add_players(database.session, players)
        add_umpire(database.session, plate_umpire)
        add_stadium(database.session, stadium)

        # This home/away determination is risky depending on ordering,
        # but every game I've looked at has worked this way.
        game["date"] = date
        game["home_team"] = teams[0]["id"]
        game["away_team"] = teams[1]["id"]
        game["stadium"] = stadium["id"]
        game["umpire_id"] = plate_umpire["id"]
        add_game(database.session, game)

        for action in actions:
            action["game_pk"] = game["game_pk"]
        database.session.add_all([Action(**action) for action in actions])

        for atbat in atbats:
            atbat["game_pk"] = game["game_pk"]
        database.session.add_all([AtBat(**atbat) for atbat in atbats])

        for pitch in pitches:
            pitch["game_pk"] = game["game_pk"]
        # HBPs don't seem to have anything in them, so skip 'em
        pitches = [p for p in pitches if p["des"] != "Hit By Pitch"]
        database.session.add_all([Pitch(**pitch) for pitch in pitches])

        database.session.commit()


def get_args():
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers()

    init_parser = subparsers.add_parser("initdb")
    init_parser.set_defaults(func=do_initdb)

    import_parser = subparsers.add_parser("import")
    import_parser.set_defaults(func=do_import)
    import_parser.add_argument("--root",
                               help="Directory to search for Gameday files")

    scraper_parser = subparsers.add_parser("scrape")
    scraper_parser.set_defaults(func=do_scrape)
    scraper_parser.add_argument("-b", "--begin", dest="begin", type=str,
                                help="Beginning date in %Y-%m-%d format")
    scraper_parser.add_argument("-e", "--end", dest="end", type=str,
                                help="Ending date in %Y-%m-%d format")
    group = scraper_parser.add_mutually_exclusive_group()
    group.add_argument("-d", "--download", dest="download",
                       action="store_true", default=False,
                       help="Download scraped files.")
    group.add_argument("-u", "--upload", dest="upload",
                       action="store_true", default=False,
                       help="Upload scraped files.")

    return parser.parse_args()


def main():
    args = get_args()
    args.func(args)



import configparser
import os

from sqlalchemy import create_engine
from sqlalchemy.orm import scoped_session, sessionmaker
from sqlalchemy.ext.declarative import declarative_base


def _get_engine_uri():
    path = os.getenv("GD_DATABASE_URI", False)
    if path:
        return path

    parser = configparser.ConfigParser()
    if parser.read("/usr/local/etc/gd/db.conf"):
        return parser.get("connection", "uri")

    raise Exception("No database engine available")


engine = create_engine(_get_engine_uri(), convert_unicode=True)
session = scoped_session(sessionmaker(autocommit=False,
                                      autoflush=False,
                                      bind=engine))
Base = declarative_base()
Base.query = session.query_property()


def init():
    # import all modules here that might define models so that
    # they will be registered properly on the metadata.  Otherwise
    # you will have to import them first before calling init_db()
    import gd.models  # noqa
    Base.metadata.create_all(bind=engine)
